from typing import List

from .token import Token
from .iterator import LineWrapIterator


class RemainingToken(Token):
    pass


def scan_for_remaining(tokens: List[Token], it: LineWrapIterator):
    start = it.copy()
    # Stop on whitespace, quotes, comments and dot followed by a digit.
    while it.next() and not (it.curr in ' \t\v\f\n\r' or it.curr in '"\'' or (
            it.curr == '/' and it.lookahead in '/*') or it.curr == '.' and it.lookahead.isnumeric()):
        pass
    tokens.append(RemainingToken(start, it))
    return True



import argparse
import sys
from pathlib import Path
from typing import List

from quom import Quom

try:
    from quom import __version__
except ImportError:
    __version__ = 'unknown'


def main(args: List[str]):
    parser = argparse.ArgumentParser(prog='quom', description='Single header generator for C/C++ libraries.')
    parser.add_argument('--version', action='version', version='quom {ver}'.format(ver=__version__))
    parser.add_argument('input_path', metavar='input', type=Path, help='Input file path of the main file.')
    parser.add_argument('output_path', metavar='output', type=Path,
                        help='Output file path of the generated single header file.')
    parser.add_argument('--stitch', '-s', metavar='format', type=str, default=None,
                        help='Format of the comment where the source files should be placed (e.g. // ~> stitch <~). \
                        Default: %(default)s (at the end of the main file)')
    parser.add_argument('--include_guard', '-g', metavar='format', type=str, default=None,
                        help='Regex format of the include guard. Default: %(default)s')
    parser.add_argument('--trim', '-t', action='store_true', default=True,
                        help='Reduce continuous line breaks to one. Default: %(default)s')
    parser.add_argument('--include_directory', '-I', type=Path, action='append', default=[],
                        help='Add include directories for header files.')
    parser.add_argument('--source_directory', '-S', type=str, action='append', default=['.'],
                        help='Set the source directories for source files. '
                             'Use ./ in front of a path to mark as relative to the header file.')

    args = parser.parse_args(args)

    # Transform source directories to distingue between:
    # - relative from header file (starting with dot)
    # - relative from workdir
    # - absolute path
    relative_source_directories = []
    source_directories = []
    for src in args.source_directory:
        path = Path(src)
        if src == '.' or src.startswith('./'):
            relative_source_directories.append(path)
        else:
            source_directories.append(path.resolve())

    with args.output_path.open('w+') as file:
        Quom(args.input_path, file, args.stitch, args.include_guard, args.trim, args.include_directory,
             relative_source_directories, source_directories)


def run():
    main(sys.argv[1:])


if __name__ == '__main__':
    run()



from typing import List

from .iterator import LineWrapIterator, Span
from .token import Token
from .tokenize_error import TokenizeError


class CommentToken(Token):
    def __init__(self, start, end, content_start, content_end):
        super().__init__(start, end)
        self.content_start = content_start.copy()
        self.content_end = content_end.copy()

    @property
    def content(self):
        return Span(self.content_start, self.content_end)


class CppCommentToken(CommentToken):
    pass


class CCommentToken(CommentToken):
    pass


def scan_for_comment_cpp_style(tokens: List[Token], it: LineWrapIterator):
    # C++-style comment: //
    if it.curr != '/' or it.lookahead != '/':
        return False
    it = LineWrapIterator(it)
    start = it.copy()
    it.next()
    it.next()
    content_start = it.copy()

    # Parse until line break.
    while it.curr not in '\n\r' and it.next():
        pass

    content_end = it.copy()

    tokens.append(CppCommentToken(start, it, content_start, content_end))
    return True


def scan_for_comment_c_style(tokens: List[Token], it: LineWrapIterator):
    # C-style comment: /*
    if it.curr != '/' or it.lookahead != '*':
        return False
    it = LineWrapIterator(it)
    start = it.copy()
    it.next()
    it.next()
    content_start = it.copy()

    # Parse until file end or */.
    while (it.curr != '*' or it.lookahead != '/') and it.next():
        pass

    if it.curr != '*':
        raise TokenizeError('C-style comment not terminated!', it)
    content_end = it.copy()
    it.next()
    it.next()

    tokens.append(CCommentToken(start, it, content_start, content_end))
    return True


def scan_for_comment(tokens: List[Token], it: LineWrapIterator):
    if not scan_for_comment_cpp_style(tokens, it):
        return scan_for_comment_c_style(tokens, it)
    return True



from typing import List

from .comment_tokenizer import scan_for_comment
from .iterator import LineWrapIterator, Span
from .number_tokenizer import scan_for_number
from .quote_tokenizer import scan_for_quote
from .remaining_tokenizer import scan_for_remaining, RemainingToken
from .token import Token, StartToken, EndToken
from .tokenize_error import TokenizeError
from .whitespace_tokenizer import scan_for_whitespace, LinebreakWhitespaceToken


class PreprocessorToken(Token):
    def __init__(self, start, end):
        super().__init__(start, end)
        self.preprocessor_tokens = None
        self.preprocessor_arguments_idx = None

    @property
    def preprocessor_instruction(self):
        return self.preprocessor_tokens[1:self.preprocessor_arguments_idx]

    @property
    def preprocessor_arguments(self):
        return self.preprocessor_tokens[self.preprocessor_arguments_idx:-1]


class PreprocessorIncludeToken(PreprocessorToken):
    def __init__(self, start, end, is_local_include: bool, path_start, path_end):
        super().__init__(start, end)
        self.is_local_include = is_local_include
        self.path = Span(path_start, path_end)


class PreprocessorUnknownIncludeToken(PreprocessorToken):
    pass


class PreprocessorPragmaToken(PreprocessorToken):
    pass


class PreprocessorPragmaOnceToken(PreprocessorPragmaToken):
    pass


class PreprocessorDefineToken(PreprocessorToken):
    pass


class PreprocessorIfNotDefinedToken(PreprocessorToken):
    pass


class PreprocessorEndIfToken(PreprocessorToken):
    pass


def scan_for_whitespaces_and_comments(it: LineWrapIterator, tokens: List[Token]):
    while it.curr != '\0':
        if scan_for_comment(tokens, it):
            continue
        if scan_for_whitespace(tokens, it):
            if isinstance(tokens[-1], LinebreakWhitespaceToken):
                return True
            continue
        break
    return it.curr == '\0'


def scan_for_line_end(it: LineWrapIterator, tokens: List[Token]):
    while it.curr != '\0':
        if scan_for_whitespaces_and_comments(it, tokens):
            return
        succeeded = scan_for_quote(tokens, it)
        if not succeeded:
            succeeded = scan_for_number(tokens, it)
        if not succeeded:
            scan_for_remaining(tokens, it)


def scan_for_preprocessor_include(start: LineWrapIterator, it: LineWrapIterator, tokens: List[Token]):
    if scan_for_whitespaces_and_comments(it, tokens) or it.curr != '"' and it.curr != '<':
        scan_for_line_end(it, tokens)
        return PreprocessorUnknownIncludeToken(start, it)

    it = LineWrapIterator(it)
    it.next()
    path_start = it.copy()
    is_local_include = False

    if it.prev == '"':
        is_local_include = True

        # Parse until non escaped ".
        backslashes = 0
        while it.next() and it.curr != '\n' and (it.curr != '"' or backslashes % 2 != 0):
            if it.curr == '\\':
                backslashes += 1
            else:
                backslashes = 0

        # Check if end of line is reached.
        if it.curr != '"':
            raise TokenizeError('Character sequence not terminated!', it)
    elif it.prev == '<':
        # Scan until terminating >.
        while it.next() and it.curr != '\n' and it.curr != '>':
            pass

        # Check if end of line is reached.
        if it.curr != '>':
            raise TokenizeError('Character sequence not terminated!', it)

    path_end = it.copy()
    it.next()

    scan_for_line_end(it, tokens)
    return PreprocessorIncludeToken(start, it, is_local_include, path_start, path_end)


def scan_for_preprocessor_pragma(start: LineWrapIterator, it: LineWrapIterator, tokens: List[Token]):
    if scan_for_whitespaces_and_comments(it, tokens):
        return PreprocessorPragmaToken(start, it)

    scan_for_remaining(tokens, it)
    if str(tokens[-1]) != 'once':
        return PreprocessorPragmaToken(start, it)

    if scan_for_whitespaces_and_comments(it, tokens):
        return PreprocessorPragmaOnceToken(start, it)

    scan_for_line_end(it, tokens)
    return PreprocessorPragmaToken(start, it)


def scan_for_preprocessor(tokens: List[Token], it: LineWrapIterator):
    if it.curr != '#':
        return None
    start = it.copy()
    it.next()

    preprocessor_tokens = [StartToken(start, start), RemainingToken(start, it)]
    if scan_for_whitespaces_and_comments(it, preprocessor_tokens):
        preprocessor_token = PreprocessorToken(start, it)
        preprocessor_arguments_idx = len(preprocessor_tokens)
    else:
        scan_for_remaining(preprocessor_tokens, it)
        name = str(preprocessor_tokens[-1])
        preprocessor_arguments_idx = len(preprocessor_tokens)

        if name == 'include':
            preprocessor_token = scan_for_preprocessor_include(start, it, preprocessor_tokens)
        elif name == 'pragma':
            preprocessor_token = scan_for_preprocessor_pragma(start, it, preprocessor_tokens)
        elif name == 'define':
            scan_for_line_end(it, preprocessor_tokens)
            preprocessor_token = PreprocessorDefineToken(start, it)
        elif name == 'ifndef':
            scan_for_line_end(it, preprocessor_tokens)
            preprocessor_token = PreprocessorIfNotDefinedToken(start, it)
        elif name == 'endif':
            scan_for_line_end(it, preprocessor_tokens)
            preprocessor_token = PreprocessorEndIfToken(start, it)
        else:
            scan_for_line_end(it, preprocessor_tokens)
            preprocessor_token = PreprocessorToken(start, it)

    preprocessor_tokens.append(EndToken(it, it))
    preprocessor_token.preprocessor_tokens = preprocessor_tokens
    preprocessor_token.preprocessor_arguments_idx = preprocessor_arguments_idx

    tokens.append(preprocessor_token)
    return True



# flake8: noqa Q003
# Tokenizer
from .tokenize import tokenize
from .token import Token, EmptyToken, StartToken, EndToken
from .tokenize_error import TokenizeError

# Tokens
from .comment_tokenizer import CommentToken, CppCommentToken, CCommentToken
from .number_tokenizer import NumberToken
from .preprocessor_tokenizer import PreprocessorToken, PreprocessorIncludeToken, PreprocessorUnknownIncludeToken, \
    PreprocessorPragmaToken, PreprocessorPragmaOnceToken, PreprocessorDefineToken, PreprocessorIfNotDefinedToken, \
    PreprocessorEndIfToken
from .quote_tokenizer import QuoteToken, SingleQuoteToken, DoubleQuoteToken
from .remaining_tokenizer import RemainingToken
from .whitespace_tokenizer import WhitespaceToken, WhitespaceWhitespaceToken, LinebreakWhitespaceToken



from io import StringIO
from pathlib import Path

from quom import Quom

FILE_MAIN_HPP = """\
#pragma once

#ifndef FOOBAR_HPP
#define FOOBAR_HPP

#include "foo.hpp"

#endif // FOOBAR_HPP
"""

FILE_FOO_HPP = """\
#pragma once

# /* */ ifndef /*123*/ FOOBAR_FOO_HPP
#define FOOBAR_FOO_HPP // abc

#include <iostream>

extern int foo;

#endif // FOOBAR_FOO_HPP
"""

FILE_FOO_CPP = """\
#include "foo.hpp"

#include <algorithm>

int foo = 42;"""

RESULT_NORMAL = """\
#pragma once

#ifndef FOOBAR_HPP
#define FOOBAR_HPP

# /* */ ifndef /*123*/ FOOBAR_FOO_HPP
#define FOOBAR_FOO_HPP // abc

#include <iostream>

extern int foo;

#endif // FOOBAR_FOO_HPP

#endif // FOOBAR_HPP

#include <algorithm>

int foo = 42;"""

RESULT_NORMAL_WITHOUT_TRIM = """\
#pragma once

#ifndef FOOBAR_HPP
#define FOOBAR_HPP



# /* */ ifndef /*123*/ FOOBAR_FOO_HPP
#define FOOBAR_FOO_HPP // abc

#include <iostream>

extern int foo;

#endif // FOOBAR_FOO_HPP


#endif // FOOBAR_HPP


#include <algorithm>

int foo = 42;"""


def init():
    with open('main.hpp', 'w+') as file:
        file.write(FILE_MAIN_HPP)

    with open('foo.hpp', 'w+') as file:
        file.write(FILE_FOO_HPP)

    with open('foo.cpp', 'w+') as file:
        file.write(FILE_FOO_CPP)


def test_normal(fs):
    init()

    dst = StringIO()
    Quom(Path('main.hpp'), dst)

    assert dst.getvalue() == RESULT_NORMAL


def test_normal_without_trim(fs):
    init()

    dst = StringIO()
    Quom(Path('main.hpp'), dst, trim=False)

    assert dst.getvalue() == RESULT_NORMAL_WITHOUT_TRIM


def test_without_newline_at_end(fs):
    with open('main.hpp', 'w+') as file:
        file.write('int a;')

    dst = StringIO()
    Quom(Path('main.hpp'), dst)

    assert dst.getvalue() == 'int a;'
